#ifndef _BATTERY_MONITOR_H
#define _BATTERY_MONITOR_H

#include "threadserver.h"
#include "eventserver.h"
#include "dynamixel.h"
#include "dynamixelserver_ftdi.h"
#include "mutex.h"

typedef struct
{
  unsigned short int error:4;
  unsigned short int fully_discharged:1;
  unsigned short int fully_charged:1;
  unsigned short int discharging:1;
  unsigned short int initialized:1;
  unsigned short int remaining_time_alarm:1;
  unsigned short int remaining_capacity_alarm:1;
  unsigned short int reserved2:1;
  unsigned short int terminate_discharge_alarm:1;
  unsigned short int over_temperature_alarm:1;
  unsigned short int reserved1:1;
  unsigned short int terminate_charge_alarm:1;
  unsigned short int over_charged_alarm:1;
}TBatteryStatus;

typedef struct
{
  double temperature;
  double battery_voltage;
  double battery_current;
  double battery_avg_current;
  double relative_state_of_charge;
  double absolute_state_of_charge;
  double remaining_capacity;
  double full_charge_capacity;
  double run_time_to_empty;
  double avg_time_to_empty;
  double avg_time_to_full;
  double charging_current;
  double charging_voltage;
  TBatteryStatus status;
  double design_capacity;
  double design_voltage;
  unsigned short int pack_status;
  double cell4_voltage;
  double cell3_voltage;
  double cell2_voltage;
  double cell1_voltage;
}TBatteryInfo;

typedef struct{
  std::string id;
  std::string event_id;
  int frequency_hz;
  int time_on_s;
  int time_off_s;
  int priority;
  bool (*condition_function)(TBatteryInfo *info);
}TAlarm;

class CBattery_Monitor
{
  private:
    // pc speacker sound variables
    bool sound_enabled;
    int console_fd;
    // serial device object
    CDynamixel *bat_mon;
    CDynamixelServerFTDI *dyn_server;
    // thread attributes
    CThreadServer *thread_server;
    std::string battery_monitor_thread_id;
    std::string battery_sound_thread_id;
    // event attributes
    CEventServer *event_server;
    std::string finish_thread_event_id;
    std::string finish_sound_event_id;
    // battery information
    TBatteryInfo info;
    CMutex info_access;
    // alarms
    std::list<TAlarm> alarms;
    std::list<TAlarm>::iterator current_alarm;
    // signal variables
    static bool signal_finish;
  protected:
    static void *battery_monitor_thread(void *param);
    static void *battery_sound_thread(void *param);
    static void signal_handler(int signum);
    void parse_packet(unsigned char *data, int len);
  public:
    CBattery_Monitor(std::string &serial_dev,bool handle_signals=true);
    // basic battery info
    void get_battery_info(TBatteryInfo *info);
    // alarm handling
    void enable_alarm_sound(void);
    void disable_alarm_sound(void);
    bool is_alarm_sound_enabled(void);
    std::string create_alarm(std::string &name,bool (*function)(TBatteryInfo *info),int priority=-1);
    void set_alarm_sound(std::string &name,int freq_hz,int time_on_s, int time_off_s);
    void delete_alarm(std::string &name);
    // signal handling
    bool finish(void);
    friend std::ostream& operator<< (std::ostream& out,CBattery_Monitor& bat);
    ~CBattery_Monitor();
};

#endif
