#include "battery_monitor.h"
#include "exceptions.h"
#include <iostream>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/kd.h>
#include <signal.h>

#ifndef CLOCK_TICK_RATE
  #define CLOCK_TICK_RATE 1193180
#endif

bool CBattery_Monitor::signal_finish=false;

CBattery_Monitor::CBattery_Monitor(std::string &serial_num,bool handle_signals)
{
  this->dyn_server=CDynamixelServerFTDI::instance();
  try{
    // create the serial port device to communicate with the battery
    this->dyn_server->config_bus(serial_num,9600);
    this->bat_mon=this->dyn_server->get_device(0x01);
    // initialize the battery information
    memset(&this->info,0x00,sizeof(TBatteryInfo));
    // initialize sound variables
    this->sound_enabled=false;
    this->console_fd=-1;
    // signal handling
    if(handle_signals)
      if(signal((int)SIGINT,CBattery_Monitor::signal_handler)==SIG_ERR)
        throw CException(_HERE_,"Impossible to assign a signal handler to the SIGINT signal");
  }catch(CException &e){
    if(this->bat_mon!=NULL)
    {
      this->dyn_server->free_device(0x01);
      this->bat_mon=NULL;
    }
    throw;
  }
  // initialize the events
  this->event_server=CEventServer::instance();
  this->finish_thread_event_id="battery_monitor_finish_thread_event";
  this->event_server->create_event(this->finish_thread_event_id);
  this->finish_sound_event_id="battery_monitor_finish_sound_event";
  this->event_server->create_event(this->finish_sound_event_id);
  // initialize the threads
  this->thread_server=CThreadServer::instance();
  this->battery_monitor_thread_id="battery_monitor_thread";
  this->thread_server->create_thread(this->battery_monitor_thread_id);
  this->thread_server->attach_thread(this->battery_monitor_thread_id,this->battery_monitor_thread,this);
  this->thread_server->start_thread(this->battery_monitor_thread_id);
  this->battery_sound_thread_id="battery_sound_thread";
  this->thread_server->create_thread(this->battery_sound_thread_id);
  this->thread_server->attach_thread(this->battery_sound_thread_id,this->battery_sound_thread,this);
}

void *CBattery_Monitor::battery_monitor_thread(void *param)
{
  CBattery_Monitor *battery=(CBattery_Monitor *)param;
  short int data[21];
  bool end=false;
  std::list<TAlarm>::iterator iterator;
  
  while(!end)
  {
    try{
      if(battery->event_server->event_is_set(battery->finish_thread_event_id))
        end=true;
      else
      {
        /* get the data from the battery */
        battery->bat_mon->read_registers(0x20,(unsigned char *)data,42);
        /* convert the data */
        battery->info.temperature=data[0]/10.0-273.15;
        battery->info.battery_voltage=data[1]/1000.0;
        battery->info.battery_current=data[2]/1000.0;
        battery->info.battery_avg_current=data[3]/1000.0;
        battery->info.relative_state_of_charge=data[4];
        battery->info.absolute_state_of_charge=data[5];
        battery->info.remaining_capacity=data[6]/1000.0;
        battery->info.full_charge_capacity=((unsigned short int)data[7])/1000.0;
        battery->info.run_time_to_empty=data[8];
        battery->info.avg_time_to_empty=data[9];
        battery->info.avg_time_to_full=data[10];
        battery->info.charging_current=data[11]/1000.0;
        battery->info.charging_voltage=data[12]/1000.0;
        (*((unsigned short int *)&battery->info.status))=data[13];
        battery->info.design_capacity=((unsigned short int)data[14])/1000.0;
        battery->info.design_voltage=data[15]/1000.0;
        // pack status ignored
        battery->info.cell4_voltage=data[17]/1000.0;
        battery->info.cell3_voltage=data[18]/1000.0;
        battery->info.cell2_voltage=data[19]/1000.0;
        battery->info.cell1_voltage=data[20]/1000.0;

        bool active_alarm=false;
        for(iterator=battery->alarms.begin();iterator!=battery->alarms.end();iterator++)
        {
          if(iterator->condition_function(&battery->info))
          {
            active_alarm=true;
            if(!battery->event_server->event_is_set(iterator->event_id))
              battery->event_server->set_event(iterator->event_id);
            // start the thread sound
            battery->current_alarm=iterator;
            if(battery->sound_enabled)
            {
              /* start the thread only once */
              if(battery->thread_server->get_thread_state(battery->battery_sound_thread_id)==attached)
                battery->thread_server->start_thread(battery->battery_sound_thread_id); 
            }
            break;
          }
        }

        if(!active_alarm)
          battery->disable_alarm_sound();

      }
    }catch(CException &e){
      std::cout << e.what() << std::endl;
    }
  }

  pthread_exit(NULL);
}

void *CBattery_Monitor::battery_sound_thread(void *param)
{
  CBattery_Monitor *battery=(CBattery_Monitor *)param;
  bool end=false;
  int i;

  while(!end)
  {
    try{
      if(battery->event_server->event_is_set(battery->finish_sound_event_id))
        end=true;
      else
      {
        if(ioctl(battery->console_fd,KIOCSOUND,(int)(CLOCK_TICK_RATE/battery->current_alarm->frequency_hz))<0)
          std::cout << "Error while setting the desired frequency" << std::endl;
        for(i=0;i<battery->current_alarm->time_on_s;i++)
        {
          if(battery->event_server->event_is_set(battery->finish_sound_event_id))
          {
            end=true;
            break;
          }
          sleep(1);
        }
        // stop the sound
        if(ioctl(battery->console_fd,KIOCSOUND,0)<0)
          std::cout << "Error while setting the desired frequency" << std::endl;
        for(i=0;i<battery->current_alarm->time_off_s;i++)
        {
          if(battery->event_server->event_is_set(battery->finish_sound_event_id))
          {
            end=true;
            break;
          }
          sleep(1);
        }
      }
    }catch(CException &e){
      std::cout << e.what() << std::endl;
    }
  }

  pthread_exit(NULL);
}

void CBattery_Monitor::signal_handler(int signum)
{
  // signal the object to finish the execution
  CBattery_Monitor::signal_finish=true;
}

void CBattery_Monitor::get_battery_info(TBatteryInfo *info)
{
  if(info!=NULL)
    memcpy(info,&this->info,sizeof(TBatteryInfo));
}

// alarm handling
void CBattery_Monitor::enable_alarm_sound(void)
{
  if(!this->sound_enabled)
  {
    if(this->console_fd!=-1)
      close(this->console_fd);
    if((this->console_fd=open("/dev/tty0",O_WRONLY)) == -1)
      throw CException(_HERE_,"Impossible to initialize the sound system");
    this->sound_enabled=true;
  } 
}

void CBattery_Monitor::disable_alarm_sound(void)
{
  // stop the thread if it is running
  if(this->sound_enabled)
  {
    if(this->thread_server->get_thread_state(this->battery_sound_thread_id)==active)
    {
      this->event_server->set_event(this->finish_sound_event_id);
      this->thread_server->end_thread(this->battery_sound_thread_id);
      this->event_server->reset_event(this->finish_sound_event_id);
    }
    if(this->console_fd!=-1)
      close(this->console_fd);
    this->console_fd=-1;
    this->sound_enabled=false;
  }
}

bool CBattery_Monitor::is_alarm_sound_enabled(void)
{
  return this->sound_enabled;
}

std::string CBattery_Monitor::create_alarm(std::string &name,bool (*function)(TBatteryInfo *info),int priority)
{
  std::list<TAlarm>::iterator iterator;
  TAlarm new_alarm;

  for(iterator=this->alarms.begin();iterator!=this->alarms.end();iterator++)
  {
    if(iterator->id==name)
    {
      // handle exceptions
      throw CException(_HERE_,"An other alarm with the same name already exists.");
    }
    else if(priority!=-1 && iterator->priority==priority)
    {
      // handle exceptions
      throw CException(_HERE_,"An other alarm with the same priority already exists.");
    }
  }
  if(function==NULL)
  {
    // handle exceptions
    throw CException(_HERE_,"Invalid condition function.");
  }
  // initialize the new alarm
  new_alarm.id=name;
  new_alarm.event_id=name+"_event_id";
  new_alarm.priority=priority;
  new_alarm.frequency_hz=440;
  new_alarm.time_on_s=1;
  new_alarm.time_off_s=10;
  this->event_server->create_event(new_alarm.event_id);
  new_alarm.condition_function=function;
  if(this->alarms.size()>0 && priority!=-1)
  {
    for(iterator=this->alarms.begin();iterator!=this->alarms.end();iterator++)
      if(iterator->priority>priority)
      {
        this->alarms.insert(iterator,new_alarm);
        break;
      }
    if(iterator==this->alarms.end())
      this->alarms.push_back(new_alarm);
  }
  else  
    this->alarms.push_back(new_alarm);

  return new_alarm.event_id;
}

void CBattery_Monitor::set_alarm_sound(std::string &name, int freq_hz,int time_on_s, int time_off_s)
{
  std::list<TAlarm>::iterator iterator;

  for(iterator=this->alarms.begin();iterator!=this->alarms.end();iterator++)
  {
    if(iterator->id==name)
    {
      iterator->frequency_hz=freq_hz;
      iterator->time_on_s=time_on_s;
      iterator->time_off_s=time_off_s;
      return;
    } 
  }
  throw CException(_HERE_,"Alarm name not found");
}

void CBattery_Monitor::delete_alarm(std::string &name)
{
  std::list<TAlarm>::iterator iterator;

  for(iterator=this->alarms.begin();iterator!=this->alarms.end();iterator++)
  {
    if(iterator->id==name)
    {
      this->event_server->delete_event(iterator->event_id);
      this->alarms.erase(iterator);
      return;
    }
  }
  // handle exceptions
  throw CException(_HERE_,"No alarm exists with the given name.");
}

bool CBattery_Monitor::finish(void)
{
  return this->signal_finish;
}

std::ostream& operator<< (std::ostream& out,CBattery_Monitor& bat)
{
  out << "Battery information:" << std::endl;
  out << "  Temperature: " << bat.info.temperature << " ºC" << std::endl;
  out << "  Voltage: " << bat.info.battery_voltage << " V" << std::endl;
  out << "  Current: " << bat.info.battery_current << " A" << std::endl;
  out << "  Average current: " << bat.info.battery_avg_current << " A" << std::endl;
  out << "  Relative state of charge: " << bat.info.relative_state_of_charge << " %" << std::endl;
  out << "  Absolute state of charge: " << bat.info.absolute_state_of_charge << " %" << std::endl;
  out << "  Remaining capacity: " << bat.info.remaining_capacity << " Ah" << std::endl;
  out << "  Full charge capacity: " << bat.info.full_charge_capacity << " Ah" << std::endl;
  out << "  Run time to empty: " << bat.info.run_time_to_empty << " min" << std::endl;
  out << "  Average time to empty: " << bat.info.avg_time_to_empty << " min" << std::endl;
  out << "  Average time to full: " << bat.info.avg_time_to_full << " min" << std::endl;
  out << "  Charging current: " << bat.info.charging_current << " A" << std::endl;
  out << "  Charging voltage: " << bat.info.charging_voltage << " V" << std::endl;
  out << "  Battery status: " << std::endl;
  out << "    over charged alarm: " << bat.info.status.over_charged_alarm << std::endl;
  out << "    terminate_charge alarm: " << bat.info.status.terminate_charge_alarm << std::endl;
  out << "    over temperature alarm: " << bat.info.status.over_temperature_alarm << std::endl;
  out << "    terminate discharge alarm: " << bat.info.status.terminate_discharge_alarm << std::endl;
  out << "    remaining capacity alarm: " << bat.info.status.remaining_capacity_alarm << std::endl;
  out << "    remaining time alarm: " << bat.info.status.remaining_time_alarm << std::endl;
  out << "    initialized: " << bat.info.status.initialized << std::endl;
  out << "    discharging: " << bat.info.status.discharging << std::endl;
  out << "    fully charged: " << bat.info.status.fully_charged << std::endl;
  out << "    fully discharged: " << bat.info.status.fully_discharged << std::endl;
  out << "    error: " << bat.info.status.error << std::endl;
  out << "  Design capacity: " << bat.info.design_capacity << " Ah" << std::endl;
  out << "  Design voltage: " << bat.info.design_voltage << " V" << std::endl;
  out << "  cell 4 voltage: " << bat.info.cell4_voltage << " V" << std::endl;
  out << "  cell 3 voltage: " << bat.info.cell3_voltage << " V" << std::endl;
  out << "  cell 2 voltage: " << bat.info.cell2_voltage << " V" << std::endl;
  out << "  cell 1 voltage: " << bat.info.cell1_voltage << " V" << std::endl;
  
  return out;
}
 
CBattery_Monitor::~CBattery_Monitor()
{
  std::list<TAlarm>::iterator iterator;

  if(this->thread_server->get_thread_state(this->battery_monitor_thread_id)==active)
  {
    // signal the thread to terminate
    this->event_server->set_event(this->finish_thread_event_id);
    // wait for the thread to finish
    this->thread_server->end_thread(this->battery_monitor_thread_id);
  }
  this->thread_server->detach_thread(this->battery_monitor_thread_id);
  this->thread_server->delete_thread(this->battery_monitor_thread_id);
  this->battery_monitor_thread_id="";
  this->event_server->delete_event(this->finish_thread_event_id);
  this->finish_thread_event_id="";
  if(this->thread_server->get_thread_state(this->battery_sound_thread_id)==active)
  {
    this->event_server->set_event(this->finish_sound_event_id);
    this->thread_server->end_thread(this->battery_sound_thread_id);
  }
  this->thread_server->detach_thread(this->battery_sound_thread_id);
  this->thread_server->delete_thread(this->battery_sound_thread_id);
  this->battery_sound_thread_id="";
  this->event_server->delete_event(this->finish_sound_event_id);
  this->finish_sound_event_id="";
  // clear all alarms
  for(iterator=this->alarms.begin();iterator!=this->alarms.end();iterator++)
  {
    this->event_server->delete_event(iterator->event_id);
    iterator->event_id="";
    iterator->id="";
    iterator->condition_function=NULL;
  }
  this->alarms.clear();
  // close the serial port
  if(this->bat_mon!=NULL)
  {
    this->dyn_server->free_device(0x01);;
    this->bat_mon=NULL;
  }
}

